"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createRebaser = void 0;
const parse5_html_rewriting_stream_1 = __importDefault(require("parse5-html-rewriting-stream"));
const source_map_1 = require("source-map");
const events_1 = require("events");
const url_1 = require("url");
const path_1 = require("path");
const slash_1 = __importDefault(require("slash"));
const stream_1 = require("stream");
const createRebaser = (map, options = {}) => {
    const eventEmitter = new events_1.EventEmitter();
    const rebase = (html) => {
        const rewritingStream = new parse5_html_rewriting_stream_1.default();
        let { rebase } = options;
        let regions = null;
        return new Promise((resolve, reject) => {
            let data = Buffer.from('');
            const inputStream = new stream_1.Readable({
                encoding: "utf-8"
            });
            const outputStream = new stream_1.Writable({
                write(chunk, _encoding, callback) {
                    data = Buffer.concat([data, chunk]);
                    callback();
                }
            });
            outputStream.on('finish', () => {
                resolve({
                    data,
                    map
                });
            });
            inputStream
                .pipe(rewritingStream)
                .pipe(outputStream);
            const isRebasable = (url) => {
                return !(0, path_1.isAbsolute)(url.href) && (url.host === null) && ((url.hash === null) || (url.path !== null));
            };
            const getRegions = () => {
                if (!regions) {
                    const foundRegions = [];
                    let sourceMapConsumer = new source_map_1.SourceMapConsumer(JSON.parse(map.toString()));
                    let region = null;
                    let currentSource = null;
                    sourceMapConsumer.eachMapping((mapping) => {
                        let source = mapping.source;
                        if (source !== currentSource) {
                            if (region) {
                                region.endLine = mapping.generatedLine;
                                region.endColumn = mapping.generatedColumn;
                            }
                            region = {
                                source: source,
                                startLine: mapping.generatedLine,
                                startColumn: mapping.generatedColumn,
                                endLine: null,
                                endColumn: null
                            };
                            foundRegions.push(region);
                            currentSource = source;
                        }
                    }, null);
                    regions = foundRegions;
                }
                return regions;
            };
            const transformStartTag = (tag) => {
                const processTag = (tag) => {
                    const attributes = tag.attrs;
                    attributes.forEach((attribute) => {
                        switch (attribute.name) {
                            case 'href':
                            case 'src':
                                const url = (0, url_1.parse)(attribute.value);
                                if (isRebasable(url)) {
                                    const location = tag.sourceCodeLocation;
                                    let tagStartLine = location.startLine;
                                    let tagStartColumn = location.startCol - 1;
                                    let i = 0;
                                    let tagRegion = null;
                                    let regions = getRegions();
                                    while ((i < regions.length) && (tagRegion === null)) {
                                        let region = regions[i];
                                        if (((region.startLine < tagStartLine) || ((region.startLine === tagStartLine) && (region.startColumn <= tagStartColumn))) &&
                                            ((region.endLine === null) || (region.endLine > tagStartLine) ||
                                                ((region.endLine === tagStartLine) && (region.endColumn === null || (region.endColumn >= tagStartColumn))))) {
                                            tagRegion = region;
                                        }
                                        i++;
                                    }
                                    const { source } = tagRegion;
                                    const resolvedPath = path_1.posix.join((0, path_1.dirname)(source), url.pathname);
                                    const done = (rebasedPath) => {
                                        if (rebasedPath !== false) {
                                            if (!rebasedPath) {
                                                rebasedPath = resolvedPath;
                                            }
                                            attribute.value = (0, slash_1.default)((0, path_1.join)('.', rebasedPath));
                                            eventEmitter.emit('rebase', rebasedPath, resolvedPath);
                                        }
                                    };
                                    if (!rebase) {
                                        rebase = (_url, _source, done) => {
                                            done();
                                        };
                                    }
                                    rebase(source, resolvedPath, done);
                                }
                                break;
                        }
                    });
                };
                processTag(tag);
            };
            rewritingStream.on('startTag', (startTag) => {
                try {
                    transformStartTag(startTag);
                    rewritingStream.emitStartTag(startTag);
                }
                catch (error) {
                    reject(error);
                }
            });
            inputStream.push(html);
            inputStream.push(null);
        });
    };
    const rebaser = {
        rebase,
        on(event, listener) {
            eventEmitter.on(event, listener);
            return rebaser;
        }
    };
    return rebaser;
};
exports.createRebaser = createRebaser;
